/*--------------------------------------------------------------------------------------------------------------------------------
 *  Gewitter Monitor v2
 *  
 *  File "GewitterMonitor_config.ino"
 *  Includes all functions for setup and configuration
 *  
 *  Stephan Laage-Witt - February 2018
 *  Version 2.0
 *  
 *  Date: 12-Feb-2018
 */

#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <EEPROM.h>

// EEPROM addresses
#define EE_BASE 10

// display coordinates
#define DP_EDIT_BOX_HEIGHT 9

// strings in program memory
const char str_opt_title_alarm_window[] PROGMEM = "Alarm  window [10...60 min]";
const char str_opt_title_alarm_threshold[] PROGMEM = "Alarm threshold";
const char str_opt_title_trigger_threshold[] PROGMEM = "Trigger threshold [100...500]";
const char str_opt_title_display[] PROGMEM = "Display light [0...10]";
const char str_opt_title_startup_screen[] PROGMEM = "Startup screen";
const char str_opt_title_exit[] PROGMEM = "Exit";
const char str_opt_startup_1[] PROGMEM = "Flash Sum";
const char str_opt_startup_2[] PROGMEM = "Flash Sum + Trend";
const char str_opt_startup_3[] PROGMEM = "Flash Count";
const char str_opt_startup_4[] PROGMEM = "Flash Count + Trend";
const char str_opt_alarm_threshold_0[]  PROGMEM = "Avg 25";
const char str_opt_alarm_threshold_1[]  PROGMEM = "Avg 50";
const char str_opt_alarm_threshold_2[]  PROGMEM = "Avg 100";
const char str_opt_alarm_threshold_3[]  PROGMEM = "Avg 150";
const char str_opt_alarm_threshold_4[]  PROGMEM = "Avg 200";
const char str_opt_alarm_threshold_5[]  PROGMEM = "Avg 250";
const char str_opt_alarm_threshold_6[]  PROGMEM = "Avg 300";
const char str_opt_alarm_threshold_7[]  PROGMEM = "Max 200";
const char str_opt_alarm_threshold_8[]  PROGMEM = "Max 300";
const char str_opt_alarm_threshold_9[]  PROGMEM = "Max 400";
const char str_opt_alarm_threshold_10[] PROGMEM = "Max 500";
const char str_opt_alarm_threshold_11[] PROGMEM = "Max 600";
const char str_opt_alarm_threshold_12[] PROGMEM = "Max 700";

const char* const str_opt_title_table[] PROGMEM = { 
    str_opt_title_trigger_threshold, str_opt_title_alarm_window, 
    str_opt_title_alarm_threshold, str_opt_title_display,
    str_opt_title_startup_screen, str_opt_title_exit
};  
  
const char* const str_opt_startup_table[] PROGMEM = {
    str_opt_startup_1, str_opt_startup_2,
    str_opt_startup_3, str_opt_startup_4
};

const char* const str_opt_alarm_threshold_table[] PROGMEM = {
  str_opt_alarm_threshold_0, str_opt_alarm_threshold_1, str_opt_alarm_threshold_2,
  str_opt_alarm_threshold_3, str_opt_alarm_threshold_4, str_opt_alarm_threshold_5,
  str_opt_alarm_threshold_6, str_opt_alarm_threshold_7, str_opt_alarm_threshold_8,
  str_opt_alarm_threshold_9, str_opt_alarm_threshold_10, str_opt_alarm_threshold_11,
  str_opt_alarm_threshold_12 
};

struct config_entry_t {
  uint16_t value;
  uint16_t val_min;
  uint16_t val_max;
  uint16_t val_step;
  uint16_t val_default;
  uint8_t dp_x_title;
  uint8_t dp_x_val_start;
  uint8_t dp_x_val;
  uint8_t dp_x_val_end;
  uint8_t dp_y;
};

config_entry_t config_table[6] = {
//    min  max step deflt 
  {0, 100, 500,  50, 200, 10, 169, 189, 190,  1},  // trigger threshold
  {0,  10,  60,   5,  15, 10, 176, 189, 190, 12},  // alarm window
  {0,   0,  12,   1,   2, 10, 148, 189, 190, 23},  // alarm threshold
  {0,   0,  10,   1,   8, 10, 176, 189, 190, 34},  // display light
  {0,   0,   3,   1,   1, 10, 100, 189, 190, 45},  // startup screen
  {0,   0,   0,   0,   0, 10,   0,   0,   0, 56}   // exit
};

/* run_setup ------------------------------------------------------------------------------------------------------------------------------
 *  Runs the setup and configuration menu
 */
void run_setup(void) {
  char str_buf[20];
  uint8_t opt = 0;
  boolean exit_flag = false;
  boolean change_flag = false;

  keyboard_value = 0;

  // initiate config_table
  get_config();
 
  // display current data
  gd.set_font(4);
  gd.clear_screen();
  for (opt = 0; opt < 6; ++opt) {
    gd.set_cursor(config_table[opt].dp_x_title, config_table[opt].dp_y);
    strcpy_P(str_buf, (char*)pgm_read_word(&(str_opt_title_table[opt])));
    gd.draw_str(str_buf);
    show_config_value(opt);
  };

  opt = 0;
  draw_arrow(config_table[opt].dp_y + 4, 1);
  do {
    if (keyboard_value > 0) {
      draw_arrow(config_table[opt].dp_y + 4, 0);
      switch (keyboard_value) {
        case KEY_SELECT:
          if (opt < 5) 
            change_flag |= edit_config_item(opt);
          else 
            exit_flag = true;
          break;
        case KEY_UP:
          if (opt == 0) opt = 5;
          else --opt;
          break;
       case KEY_DOWN:
          if (++opt == 6) opt = 0;
          break;
      };
      draw_arrow(config_table[opt].dp_y + 4, 1);
      keyboard_value = 0;
    } else {
      sleep_mode();
    };
  } while (!exit_flag);

  // save new config data
  if (change_flag) {
    write_config_to_eeprom();
    graph_mode = startup_screen;
  };
}

/* edit_config_item ------------------------------------------------------------------------------------------------------------------------
 *  
 */
boolean edit_config_item(uint8_t opt) {
  boolean exit_flag = false;
  boolean change_flag = false;
  char str_buf[25];

  keyboard_value = 0;
  show_config_box(opt, 1);
  do {
    if (keyboard_value > 0) {
      switch (keyboard_value) {
        case KEY_SELECT:
          exit_flag = true;
          break;
        case KEY_UP: 
          if (config_table[opt].value < config_table[opt].val_max) 
            config_table[opt].value += config_table[opt].val_step;
          else
            config_table[opt].value = config_table[opt].val_min;
          change_flag = true;
          break;
        case KEY_DOWN: 
          if (config_table[opt].value > config_table[opt].val_min) 
            config_table[opt].value -= config_table[opt].val_step;
          else   
            config_table[opt].value = config_table[opt].val_max;
          change_flag = true;
          break;
      };
      keyboard_value = 0;
      show_config_value(opt);
      if (opt == 3) gd.set_light(config_table[opt].value);
    } else {
      sleep_mode();
    };
  } while (!exit_flag);
  
  show_config_box(opt, 0);
  return(change_flag);
}

/* show_config_value -----------------------------------------------------------------------------------------------------------------------
 *  Prints a configuartion value in order to show actual or update configuration data
 */
void show_config_value(uint8_t opt) {
  char str_buf[25];
  uint8_t str_length;

  if (opt > 4) return;
  gd.draw_filled_rectangle(config_table[opt].dp_x_val_start, 
    config_table[opt].dp_y + 1, 
    config_table[opt].dp_x_val_end, 
    config_table[opt].dp_y + DP_EDIT_BOX_HEIGHT - 1, 0);
  gd.set_cursor(config_table[opt].dp_x_val, config_table[opt].dp_y);
  if (opt == 2)                                                   // option "alarm threshold" 
    strcpy_P(str_buf, (char*)pgm_read_word(&(str_opt_alarm_threshold_table[config_table[opt].value])));
  else if (opt == 4)                                              // options "startup screen"
    strcpy_P(str_buf, (char*)pgm_read_word(&(str_opt_startup_table[config_table[opt].value])));
  else 
    itoa(config_table[opt].value, str_buf, 10);
  gd.draw_radj_str(str_buf);
}

/* show_config_box -------------------------------------------------------------------------------------------------------------------------
 *  Draws a box around the value of a config item
 */
void show_config_box(uint8_t opt, uint8_t mode) {
  if (opt > 4) return;
  gd.draw_rectangle(config_table[opt].dp_x_val_start - 1, 
    config_table[opt].dp_y - 1, 
    config_table[opt].dp_x_val_end + 1, 
    config_table[opt].dp_y + DP_EDIT_BOX_HEIGHT + 1,
    mode);
};
 
/* draw_arrow ------------------------------------------------------------------------------------------------------------------------------
 *  
 */
 void draw_arrow(uint8_t dp_y, uint8_t mode) {
   gd.draw_line(0, dp_y, 7, dp_y, mode);
   gd.draw_line(4, dp_y - 3, 7, dp_y, mode);
   gd.draw_line(4, dp_y + 3, 7, dp_y, mode);
 };

 /* get_config_from_eeprom -----------------------------------------------------------------------------------------------------------------
  *  Reads the data of the configuration table from the eeprom
  */
void get_config_from_eeprom(void) {
  int eeprom_adr = EE_BASE;
  
  for (uint8_t i = 0; i < 5; ++i) {
    config_table[i].value = (EEPROM.read(eeprom_adr) << 8) + EEPROM.read(eeprom_adr + 1);
    if ((config_table[i].value > config_table[i].val_max) ||
        (config_table[i].value < config_table[i].val_min)) {
      config_table[i].value = config_table[i].val_default;
      EEPROM.write(eeprom_adr, config_table[i].value >> 8);
      EEPROM.write(eeprom_adr + 1 , config_table[i].value & 0x00ff);
    };
    eeprom_adr += 2;
  };
  write_config();
  graph_mode = startup_screen;
};    

/* write_config_to_eeprom -----------------------------------------------------------------------------------------------------------------
  *  Reads the data of the configuration table from the eeprom
  */
void write_config_to_eeprom(void) {
  int eeprom_adr = EE_BASE;
  
  for (uint8_t i = 0; i < 5; ++i) {
    EEPROM.update(eeprom_adr, config_table[i].value >> 8);
    EEPROM.update(eeprom_adr + 1, config_table[i].value & 0x00ff);
    eeprom_adr += 2;
  };
  write_config();
};    

/* get_config -----------------------------------------------------------------------------------------------------------------------------
 *  Copies the configuration data into the config table
 */
void get_config(void) {
  config_table[0].value = trigger_threshold; 
  config_table[1].value = alarm_window; 
  config_table[2].value = alarm_threshold; 
  config_table[3].value = display_light;
  config_table[4].value = startup_screen; 
}

/* write_config ---------------------------------------------------------------------------------------------------------------------------
 *  Copies the data from the configuration table to the configuration parameters
 */
void write_config(void) {
  trigger_threshold = config_table[0].value; 
  alarm_window = config_table[1].value; 
  alarm_threshold = config_table[2].value; 
  display_light = config_table[3].value; 
  startup_screen = config_table[4].value; 
}


