/*--------------------------------------------------------------------------------------------------------------------------------
 *  Gewitter Monitor v2
 *  
 *  File "GewitterMonitor_graphs.ino"
 *  Includes all graphic display functions
 *  
 *  Stephan Laage-Witt - February 2018
 *  Version 2.0
 *  
 *  Date: 12-Feb-2018
 */

#include <Wire.h>
#include <glcd_functions_avr.h>
#include <avr/pgmspace.h>

// display coordinates
#define DP_GRAPH_TITLE_X 0
#define DP_GRAPH_TITLE_Y 0
#define DP_GRAPH_MAX_X 0
#define DP_GRAPH_MAX_Y 11
#define DP_GRAPH_AVG_X 0
#define DP_GRAPH_AVG_Y 19
#define DP_GRAPH_TREND_X 0
#define DP_GRAPH_TREND_Y 27
#define DP_GRAPH_TITLE_BOX_X_END 48

// constants
#define REGRESSION_WINDOW 45
#define SCALE_WINDOW 60

// strings in program memory
const char str_sys_title[] PROGMEM = "Gewitter Monitor v2";
const char str_version[] PROGMEM = "07-Aug-2018 - www.laagewitt.de";
const char str_flash_sum_title[] PROGMEM = "Flash Sum";
const char str_flash_cnt_title[] PROGMEM = "Flash Count";
const char str_max_label[] PROGMEM = "Max";
const char str_avg_label[] PROGMEM = "Avg";
const char str_trend_label[] PROGMEM = "Trend";
const char str_empty[] PROGMEM = "---.-";

// show title --------------------------------------------------------------------------------------------------------------------
void show_title(void) {
  uint8_t font_height;
  const uint8_t title_x_pos = 29, title_y_pos = 25, title_len = 134;

  gd.clear_screen();                          
  gd.set_light(display_light);
  gd.set_font(6);                             
  delay(20);
  font_height = gd.get_font_height();
  gd.set_cursor(192/2, title_y_pos);                        
  gd.draw_center_pgm_str(str_sys_title); 
  gd.set_font(0);
  gd.set_cursor(192/2, 56);
  gd.draw_center_pgm_str(str_version);
  gd.draw_line(0, title_y_pos - 8, title_x_pos + title_len - 4, title_y_pos - 8, 1);
  gd.draw_line(0, title_y_pos - 6, title_x_pos + title_len - 2, title_y_pos - 6, 1);
  gd.draw_line(0, title_y_pos - 4, title_x_pos + title_len, title_y_pos - 4, 1);
  gd.draw_line(title_x_pos, title_y_pos + font_height +  2, 191, title_y_pos + font_height + 2, 1);
  gd.draw_line(title_x_pos + 2, title_y_pos + font_height + 4, 191, title_y_pos + font_height + 4, 1);
  gd.draw_line(title_x_pos + 4, title_y_pos + font_height + 6, 191, title_y_pos + font_height + 6, 1);
}

// find_buffer_max ---------------------------------------------------------------------------------------------------------------
uint16_t find_buffer_max(uint8_t graph_mode, uint8_t scale_window) {
  uint8_t my_buf_pnt = buffer_pnt;
  uint16_t my_max = 0;

  for (uint8_t i = 0; i < scale_window; ++i) {
    if (graph_mode < 2) {
      if (flash_sum_buffer[my_buf_pnt] > my_max)
        my_max = flash_sum_buffer[my_buf_pnt];
    } else {
      if (flash_cnt_buffer[my_buf_pnt] > my_max)
        my_max = flash_cnt_buffer[my_buf_pnt];
    };
    if (my_buf_pnt == 0)
      my_buf_pnt = BUFFER_SIZE - 1;
    else
      --my_buf_pnt;
  };
  return(my_max);
}

// draw_graph --------------------------------------------------------------------------------------------------------------------
void draw_graph(uint16_t *max_flash_sum, uint16_t *avg_flash_sum) {
  uint16_t max_flash_cnt, avg_flash_cnt; 
  uint16_t buffer_max;
  int32_t b;                                    // paramaters for linear regression 
  float m;                                      // slop for linear regression 
  char str_buf[6];
  uint8_t scale_index;
  boolean overflow_flag, regression_flag;
  struct scale_factor_t {
    uint16_t scale_max;                         // maximum value of the y-scale
    uint8_t tick_mark;                          // step between tick-marks
  };
  const scale_factor_t scale_table[11] = {
    {10, 2},
    {25, 5},
    {50, 1},
    {100, 2},
    {250, 5},
    {500, 1},
    {1000, 2},
    {2500, 5},
    {5000, 1},
    {10000, 2},
    {25000, 5}, 
  };

  // clear screen and draw static elements
  gd.clear_screen();
  gd.set_font(2);
  // gd.draw_line(BUFFER_SIZE, 0, BUFFER_SIZE, 63, 1);
  gd.set_cursor(DP_SEC_LABEL_X, DP_SEC_LABEL_Y);
  gd.draw_str("Sec");
  gd.draw_line(DP_SEC_LABEL_X, DP_SEC_Y + 9,191, DP_SEC_Y + 9, 1);  
  gd.set_cursor(DP_FLASH_CNT_LABEL_X, DP_FLASH_CNT_LABEL_Y);
  gd.draw_str("Count");
  gd.draw_line(DP_FLASH_CNT_LABEL_X, DP_FLASH_CNT_Y + 9, 191, DP_FLASH_CNT_Y + 9, 1);  
  gd.set_cursor(DP_FLASH_SUM_LABEL_X, DP_FLASH_SUM_LABEL_Y);
  gd.draw_str("Sum");
  gd.draw_line(DP_FLASH_SUM_LABEL_X, DP_FLASH_SUM_Y + 9, 191, DP_FLASH_SUM_Y + 9, 1);  
  delay(5);

  // draw timeline (x-axis) with tick marks
  for (uint8_t i = 0; i < BUFFER_SIZE - DP_GRAPH_TITLE_BOX_X_END; i += 15) {
    if ((i == 0) || (i == 60)) {
      gd.draw_line(BUFFER_SIZE - i, 0, BUFFER_SIZE - i, 5, 1);    
    } else {
      gd.set_pixel(BUFFER_SIZE - i, 0, 1);    
      gd.set_pixel(BUFFER_SIZE - i, 1, 1); 
    };
  };

  // draw border line for alarm window
  gd.draw_dotted_ver_line(BUFFER_SIZE - alarm_window, 0, 63, 5, 1);
  
  // find maximum in last SCALE_WINDOW to calculate the sacle 
  buffer_max = find_buffer_max(graph_mode, SCALE_WINDOW);

  // calculate scale for y-axis based on the scale_table
  for (scale_index = 0; scale_index < sizeof(scale_table) / sizeof(scale_factor_t); ++scale_index) {
    if (buffer_max < scale_table[scale_index].scale_max) 
      break;
  };
  if (scale_index >= sizeof(scale_table) / sizeof(scale_factor_t)) {      // if scale_max exceeds last entry in the table ...
    scale_index = (sizeof(scale_table) / sizeof(scale_factor_t)) - 1;     // ... then reset to last entry
    overflow_flag = true;
  } else {
    overflow_flag = false;
  };
  
  // draw y-axis tickmarks and labels
  gd.set_font(1);
  gd.set_pixel(BUFFER_SIZE + 1, 63, 1);
  gd.set_pixel(BUFFER_SIZE + 2, 63, 1);
  gd.set_cursor(BUFFER_SIZE + 4, 63 - 4);
  gd.draw_udec(0, 2, 0); 
  for (uint8_t i = 1; i < 5; ++i) {
    gd.set_pixel(BUFFER_SIZE + 1, 64 - (i * 13), 1);
    gd.set_pixel(BUFFER_SIZE + 2, 64 - (i * 13), 1);
    gd.set_cursor(BUFFER_SIZE + 4, 64 - (i * 13) -2);
    gd.draw_udec(i * scale_table[scale_index].tick_mark, 2, 0);
  };
  gd.set_pixel(BUFFER_SIZE + 1, 0, 1);
  gd.set_pixel(BUFFER_SIZE + 2, 0, 1);
  gd.set_cursor(BUFFER_SIZE + 4, 0);
  itoa(scale_table[scale_index].scale_max, str_buf, 10);
  if (scale_table[scale_index].tick_mark < 5)
    gd.draw_char(' ');
  gd.draw_str(str_buf);  

  // draw curve across entire timeline
  uint8_t my_buf_pnt = buffer_pnt + 1;                    // set my_buf_pnt to the start of timeline (left end of graph)
  if (my_buf_pnt == BUFFER_SIZE) my_buf_pnt = 0;
  for (uint8_t i = 0; i < BUFFER_SIZE; ++i) {             // run through the entire buffer
    uint16_t y;
    if (graph_mode < 2) 
      y = flash_sum_buffer[my_buf_pnt];
    else 
      y = flash_cnt_buffer[my_buf_pnt];
    if (y > 0) {                                          // calculate y coordinate and draw vertical line
      if (y > scale_table[scale_index].scale_max) {       // check whether we are beyond the scale
        gd.draw_line(i, 63, i, 0, 1);                     // in that case, draw line to maximum
      } else {                                            // otherwise draw line to corresponding display value
        int y_pos = 63 - ((uint32_t) y * 64 / scale_table[scale_index].scale_max);
        gd.draw_line(i, 63, i, y_pos, 1);
      };
    };
    if (++my_buf_pnt == BUFFER_SIZE)                      // advance my_buf_pnt by 1 minute
      my_buf_pnt = 0;
  };

  // calculate regression
  if ((graph_mode == 1) || (graph_mode == 3)) {
    if (graph_mode == 1) 
      regression_flag = calc_lin_reg(flash_sum_buffer, REGRESSION_WINDOW, &m, &b);
    else
      regression_flag = calc_lin_reg(flash_cnt_buffer, REGRESSION_WINDOW, &m, &b);
    if (abs(m) < 0.1) regression_flag = false;                // stop here if trend (m) is too close to 0 
    if (regression_flag) {

      int32_t x1, y1, x2, y2;
      x1 = 0;                                                 // clipping at left border (left side of regression window)
      y1 = b;
      if (y1 < 0) {
        x1 = (int32_t) -b / m;
        y1 = 0;
      } else if (y1 > scale_table[scale_index].scale_max) {
        x1 = (int32_t) (scale_table[scale_index].scale_max - b) / m;
        y1 = scale_table[scale_index].scale_max;
      };
      x2 = REGRESSION_WINDOW;                                 // clipping at right border (right side of regression window)
      y2 = (int32_t) m * x2 + b;
      if (y2 < 0) {
        x2 = (int32_t) -b / m;
        y2 = 0; 
      } else if (y2 > scale_table[scale_index].scale_max) {
        x2 = (int32_t) (scale_table[scale_index].scale_max - b) / m;
        y2 = scale_table[scale_index].scale_max;
      };

      gd.draw_dotted_ver_line(BUFFER_SIZE - REGRESSION_WINDOW, 0, 63, 5, 1);    // mark regression time window on display
      uint32_t y_start = 64 - (long) y1 * 64 / scale_table[scale_index].scale_max;
      uint32_t y_end = 64 - (long) y2 * 64 / scale_table[scale_index].scale_max;
      gd.draw_line(BUFFER_SIZE - 1 - REGRESSION_WINDOW + x1, y_start,                    // draw regression line
        BUFFER_SIZE - 1 - REGRESSION_WINDOW + x2, y_end, 1);
    };
  };

  // calculate max and average for alarm window
  calc_sum_max_avg(max_flash_sum, avg_flash_sum);
  calc_cnt_max_avg(&max_flash_cnt, &avg_flash_cnt);
  
  // print title, max, avg and trend values -------
  // draw box
  uint8_t y_size;                                                           
  if ((graph_mode == 1) || (graph_mode == 3)) y_size = DP_GRAPH_TREND_Y + 8;
  else y_size = DP_GRAPH_AVG_Y + 8;
  gd.draw_filled_rectangle(0, 0, DP_GRAPH_TITLE_BOX_X_END + 1, y_size - 1, 0);

  // print graph title
  gd.set_font(4);
  gd.set_cursor(DP_GRAPH_TITLE_X, DP_GRAPH_TITLE_Y);                         
  if (graph_mode < 2) gd.draw_pgm_str(str_flash_sum_title);
  else gd.draw_pgm_str(str_flash_cnt_title);

  // print maximum
  uint16_t show_max;
  gd.set_font(2);
  gd.set_cursor(DP_GRAPH_MAX_X, DP_GRAPH_MAX_Y);                         
  gd.draw_pgm_str(str_max_label);
  gd.set_cursor(DP_GRAPH_TITLE_BOX_X_END, DP_GRAPH_MAX_Y);
  if (graph_mode < 2) show_max = *max_flash_sum;
  else show_max = max_flash_cnt;
  if (show_max < UPPER_LIMIT) 
    itoa(show_max, str_buf, 10);
  else 
    strcpy(str_buf, "> 30k");
  gd.draw_radj_str(str_buf);

  // print average
  uint16_t show_avg;
  gd.set_cursor(DP_GRAPH_AVG_X, DP_GRAPH_AVG_Y);                          
  gd.draw_pgm_str(str_avg_label);
  gd.set_cursor(DP_GRAPH_TITLE_BOX_X_END, DP_GRAPH_AVG_Y);
  if (graph_mode < 2) show_avg = *avg_flash_sum;
  else show_avg = avg_flash_cnt;
  itoa(show_avg, str_buf, 10);
  gd.draw_radj_str(str_buf);

  // print trend
  if ((graph_mode == 1) || (graph_mode == 3)) {                           
    gd.set_cursor(DP_GRAPH_TREND_X, DP_GRAPH_TREND_Y);
    gd.draw_pgm_str(str_trend_label);
    if (regression_flag) {
      trend_to_str(m, str_buf);                                   // print trend
      gd.set_cursor(DP_GRAPH_TITLE_BOX_X_END + 4, 2);             // print arrows, up or down
      gd.set_font(7);
      if (graph_mode == 1) {                                      // trend for flash sum
        if (m > 5.0) gd.draw_char(34);
        if (m > 50.0) gd.draw_char(34);
        if (m < -5.0) gd.draw_char(38);
        if (m < -50.0) gd.draw_char(38);
      } else {                                                    // trend for flash count
        if (m > 0.2) gd.draw_char(34);
        if (m > 2.0) gd.draw_char(34);
        if (m < -0.2) gd.draw_char(38);
        if (m < -2.0) gd.draw_char(38);   
      };
      gd.set_font(2);
    } else {
      strcpy(str_buf, "--.-");
    };
    gd.set_cursor(DP_GRAPH_TITLE_BOX_X_END, DP_GRAPH_TREND_Y);
    gd.draw_radj_str(str_buf);
  };

  gd.draw_line(0, y_size, DP_GRAPH_TITLE_BOX_X_END + 1, y_size, 1);
  gd.draw_line(DP_GRAPH_TITLE_BOX_X_END + 2, y_size, DP_GRAPH_TITLE_BOX_X_END + 2, 0, 1);
}

/* trend_to_str ------------------------------------------------------------------------------------------------------------------
 *  Converst the trend (m factor of linear regression) to a string  
 *  Factor is divided by 10 and shown with a decimal fraction of 10
 *  '+' or '-' are included, always
 *  
 */
void trend_to_str(float m, char str_buf[]) {
  uint8_t digit;
  uint8_t i = 0;
  int32_t trend;
  
  if (m > 0) trend = (uint32_t) (m * 10 + 0.5);
  else trend = (uint32_t) (m * 10 - 0.5);
  if (trend == 0) {
    str_buf[i++] = '0';
    str_buf[i] = '\0';
    return;
  };
  
  if (trend > 0) str_buf[i++] = '+';
  else {
    str_buf[i++] = '-';
    trend = -trend;
  };
  digit = trend / 1000;
  if (digit > 0) {
    str_buf[i++] = '0' + digit;
    trend -= 1000 * digit;
  };
  digit = trend / 100;
  if (digit > 0) {
    str_buf[i++] = '0' + digit;
    trend -= 100 * digit;
  };
  digit = trend / 10;
  str_buf[i++] = '0' + digit;
  if (digit > 0) trend -= 10 * digit;
  str_buf[i++] = '.';
  str_buf[i++] = '0' + trend;
  str_buf[i] = '\0';
}




