/*--------------------------------------------------------------------------------------------------------------------------------
 *  Gewitter Monitor v2
 *  
 *  File "GewitterMonitor_stats.ino"
 *  Includes all statistics functions 
 *  
 *  Stephan Laage-Witt - February 2018
 *  Version 2.0
 *  
 *  Date: 12-Feb-2018
 */

#include <math.h>

/* calc_sum_max_avg --------------------------------------------------------------------------------------------------------------
 * Calculate the maximum and average of the flash sum buffer
 * Reading global variable: alarm_window, buffer_pnt
*/
void calc_sum_max_avg(uint16_t *max_flash_sum, uint16_t *avg_flash_sum) {
  uint32_t sum;
  uint8_t my_buf_pnt = buffer_pnt;
  
  *max_flash_sum = 0;
  sum = 0;
  for (uint8_t i = 0; i < alarm_window; ++i) {
    if (flash_sum_buffer[my_buf_pnt] > *max_flash_sum)       // find maximum
      *max_flash_sum = flash_sum_buffer[my_buf_pnt];
    sum += flash_sum_buffer[my_buf_pnt];                    // sum up data
    if (my_buf_pnt == 0) my_buf_pnt = BUFFER_SIZE - 1;      // next data point (count my_buf_pnt backwards)
    else --my_buf_pnt;
  };
  *avg_flash_sum = sum / alarm_window;                       // calculate average
}

/* calc_cnt_max_avg --------------------------------------------------------------------------------------------------------------
 * Calculate the maximum and average of the flash sum buffer
 * Reading global variables: alarm_window, buffer_pnt
*/
void calc_cnt_max_avg(uint16_t *max_flash_cnt, uint16_t *avg_flash_cnt) {
  uint32_t sum;
  uint8_t my_buf_pnt = buffer_pnt;
  
  *max_flash_cnt = 0;
  sum = 0;
  for (uint8_t i = 0; i < alarm_window; ++i) {
    if (flash_cnt_buffer[my_buf_pnt] > *max_flash_cnt)       // find maximum
      *max_flash_cnt = flash_cnt_buffer[my_buf_pnt];
    sum += flash_cnt_buffer[my_buf_pnt];                    // sum up data
    if (my_buf_pnt == 0) my_buf_pnt = BUFFER_SIZE - 1;      // next data point (count my_buf_pnt backwards)
    else --my_buf_pnt;
  };
  *avg_flash_cnt = sum / alarm_window;                       // calculate average
}


/* calc_alarm_level --------------------------------------------------------------------------------------------------------------
 *  Calculates alarm level based on maximum and average
 *  Algorithm to be used is defined by configuration "alarm_threshold"
 *  Reading global variables: alarm_threshold
 */
uint8_t calc_alarm_level(uint16_t max_flash_sum, uint16_t avg_flash_sum) {
  const uint16_t threshold[] = {25, 50, 100, 150, 200, 250, 300, 200, 300, 400, 500, 600, 700};
  uint16_t my_threshold;

  my_threshold = threshold[alarm_threshold];                        // select relevant threshold
  
  if (alarm_threshold > 6) {                                        // algorithm for alarm level based on maximum 
    if (max_flash_sum < my_threshold) return(0);                    // return current alarm level ...
    if (max_flash_sum < 2 * my_threshold) return(1);                // ... based on maximum value
    if (max_flash_sum < 4 * my_threshold) return(2);
    if (max_flash_sum < 8 * my_threshold) return(3);
    return(4);

  } else {
    
    if (avg_flash_sum < my_threshold) return(0);                    // algorithm for alarm levels based on average 
    if (avg_flash_sum < 2 * my_threshold) return(1);                // return current alarm level ...
    if (avg_flash_sum < 4 * my_threshold) return(2);                // ... based on average value
    if (avg_flash_sum < 8 * my_threshold) return(3);
    return(4);
  };
}

/* calc_lin_reg ------------------------------------------------------------------------------------------------------------------
 *  Calculates a linear regression for buffer_values within the specified time window
 *  Checks validity and correlation coefficient. 
 *  Return value: true -> lin regression successful
 *                false -> lin regrssion failed due to lack of correlation
 */
boolean calc_lin_reg(uint16_t *data_buffer, uint8_t minute_cnt, float *m, int32_t *b) {
  uint32_t x_sum = 0, y_sum = 0, x_square_sum = 0;
  float y_square_sum = 0.0, x_y_sum = 0.0;
  float sxx, sxy, syy, x_avg, y_avg, denominator, r;
  uint8_t my_buf_pnt = buffer_pnt;
  
  for (uint8_t i = 0; i < minute_cnt; ++i) {        
    x_sum += minute_cnt - i;                              // add up sum values
    x_square_sum += (minute_cnt - i) * (minute_cnt - i);
    y_sum += data_buffer[my_buf_pnt]; 
    y_square_sum += (float) data_buffer[my_buf_pnt] * data_buffer[my_buf_pnt];
    x_y_sum += (float) (minute_cnt - i) * data_buffer[my_buf_pnt];
    if (my_buf_pnt == 0)                                  // count my_buf_pnt backwards
      my_buf_pnt = BUFFER_SIZE - 1;
    else
      --my_buf_pnt;
  };

  // calculate regression interim values
  x_avg = (float) x_sum / minute_cnt;
  y_avg = (float) y_sum / minute_cnt;
  sxx = (float) x_square_sum - (minute_cnt * x_avg * x_avg);
  sxy = x_y_sum - (minute_cnt * x_avg * y_avg);
  syy = y_square_sum - (minute_cnt * y_avg * y_avg);
  denominator = sxx * syy;

  // check validity and calcuate correlation coefficient
  if (denominator <= 0) return(false);                  // stop in case of negative denominator
  r = sxy / sqrt(sxx * syy);
  if (abs(r) < 0.25) return(false);                     // stop in case of correlation below 0.25

  // calculate regression coefficients
  *m = sxy / sxx;                                       
  *b = (int32_t) y_avg - *m * x_avg;
  return(true);                                         // return regression success
}

/* calc_alarm_level --------------------------------------------------------------------------------------------------------------
 * Calculates the alarm level based on the flash sum. Alarm levels are defined as multiplies (*8, *4 *2, *1) of
 * the global variable alarm_threshold
 * Return: alarm_level (0 ... 4)
 */
uint8_t calc_alarm_level(uint16_t flash_sum) {
  if (flash_sum > 8 * alarm_threshold) return(4);       // > 2000
  if (flash_sum > 4 * alarm_threshold) return(3);       // > 1000
  if (flash_sum > 2 * alarm_threshold) return(2);       // > 500
  if (flash_sum > alarm_threshold) return(1);           // > 250
  return(0);
}

