  /*--------------------------------------------------------------------------------------------------------------------------------
 *  Gewitter Monitor v2
 *  
 *  Stephan Laage-Witt - February 2018
 *  Version 2.0
 *  
 *  Target board: Arduino Nano
 *  
 *  Date: 12-Feb-2018
 *  Last update: 24-Feb-2018
 */

#include <Wire.h>
#include <glcd_functions_avr.h>
#include <avr/sleep.h>
#include <avr/pgmspace.h>

// port definitions
#define PIN_KEY_SELECT 2
#define PIN_KEY_DOWN 3
#define PIN_KEY_UP 4
#define PIN_ALARM_2 6
#define PIN_ALARM_3 7
#define PIN_ALARM_4 8
#define PIN_FLASH_0 9
#define PIN_FLASH_1 10

// Key definitions
#define KEY_SELECT 1
#define KEY_DOWN 2
#define KEY_UP 3
#define KEY_CONFIG 10

// display coordinates
#define DP_SEC_LABEL_X 167
#define DP_SEC_LABEL_Y 58
#define DP_SEC_X 191
#define DP_SEC_Y 58
#define DP_FLASH_CNT_LABEL_X 167
#define DP_FLASH_CNT_LABEL_Y 13
#define DP_FLASH_CNT_X 191
#define DP_FLASH_CNT_Y 22
#define DP_FLASH_SUM_LABEL_X 167
#define DP_FLASH_SUM_LABEL_Y 35
#define DP_FLASH_SUM_X 191
#define DP_FLASH_SUM_Y 44

// job flags
#define JF_SEC 0                          // job flag "every second"

// other constants
#define IRQ_FREQ 4000                     // number of interrupts per second
#define BUFFER_SIZE 151                   // number of data points in buffer (150 minutes)
#define DISPLAY_TIMEOUT 120               // time (seconds) till display light shutoff (2 minutes)
#define FLASH_LED_DUR IRQ_FREQ/5          // duration to keep flash LEDs active (1/5 secs) 
#define UPPER_LIMIT 30000                 // upper limit for flash sum

// Configuration parameters ------------------------------------------------------------------------------------------------------
uint16_t trigger_threshold;               // treshold for detection of flashes
uint8_t alarm_window;                     // time window [min] for alarm assessment
uint16_t alarm_threshold;                 // alarm threshold levels
uint8_t display_light;                    // display light level
uint8_t startup_screen;                   // the screen shown after power on

// Global varibales ---------------------------------------------------------------------------------------------------------------
volatile uint8_t job_flags = 0;           // job flags are set by the interrupt routine and evaluated by the main program             
volatile uint16_t sec;                    // counter for seconds
volatile uint8_t keyboard_value = 0;      // current keyboard value
volatile uint16_t flash_cnt = 0;          // number of flashes per running minute 
volatile uint16_t flash_sum = 0;          // sum of flash values per running minute
volatile uint8_t display_cnt = DISPLAY_TIMEOUT;     // remaining time until display light shutoff
volatile uint8_t buffer_pnt;              // pointer for flash_sum and flash_cnt buffers
uint8_t display_status = 1;               // flag to capture current display light status (0 -> off, 1 -> on)
uint8_t alarm_level = 0;                  // current alarm level, range 0 ... 4
uint8_t graph_mode;                       // 0 -> sum, 1 -> sum + regression, 2 -> count, 3 -> count + regression

volatile uint16_t flash_sum_buffer[BUFFER_SIZE];
volatile uint16_t flash_cnt_buffer[BUFFER_SIZE];

// graphic display
glcd gd(0x20);                            // instance of the graphic display at I2C address 0x20

// setup() -----------------------------------------------------------------------------------------------------------------------
void setup() {

  // get configuration data
  get_config_from_eeprom();
  
  // start display and print title 
  Wire.begin();
  show_title();
  
  // set ports for LEDs and keyboard
  pinMode(PIN_ALARM_2, OUTPUT);
  digitalWrite(PIN_ALARM_2, HIGH);
  pinMode(PIN_ALARM_3, OUTPUT);
  digitalWrite(PIN_ALARM_3, HIGH);
  pinMode(PIN_ALARM_4, OUTPUT);
  digitalWrite(PIN_ALARM_4, HIGH);
  pinMode(PIN_FLASH_0, OUTPUT);
  digitalWrite(PIN_FLASH_0, HIGH);
  pinMode(PIN_FLASH_1, OUTPUT);
  digitalWrite(PIN_FLASH_1, HIGH);
  pinMode(PIN_KEY_SELECT, INPUT_PULLUP);
  pinMode(PIN_KEY_DOWN, INPUT_PULLUP);
  pinMode(PIN_KEY_UP, INPUT_PULLUP);

  // initiate ADC
  ADCSRA = (1<<ADEN) | (1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0);  // scaling factor 128 -> ADC input clock 125kHz 
  ADMUX  = (1<<REFS0) | (1<<REFS1) | (0<<ADLAR) | 0;          // result is right adjusted, use internal reference
  ADCSRA |= (1<<ADSC);                                        // start first ADC conversion
  
  // start timer1 with cyclic interrrupts at 4000Hz
  TCNT1 = 0;
  TCCR1A = 0;
  TCCR1B = 0;
  TCCR1B |= (1 << WGM12);                                 // activate CTC mode
  TCCR1B |= (0 << CS12) | (1 << CS11) | (0 << CS10);      // set prescale to 8
  OCR1A = 500 - 1;                                        // set compare match register, interrupts at 16MHz / 8 / 1000 -> 2000Hz
  TIMSK1 |= (1 << OCIE1A);                                // enable timer inerrupts

  // flash LEDs to show they are working
  digitalWrite(PIN_ALARM_2, LOW);
  delay(500);
  digitalWrite(PIN_ALARM_3, LOW);
  delay(500);
  digitalWrite(PIN_ALARM_4, LOW);
  delay(500);
  digitalWrite(PIN_ALARM_3, HIGH);
  digitalWrite(PIN_ALARM_2, HIGH);
  digitalWrite(PIN_ALARM_4, HIGH);
  digitalWrite(PIN_FLASH_1, LOW);
  delay(500);
  digitalWrite(PIN_FLASH_1, HIGH);
  digitalWrite(PIN_FLASH_0, LOW);
  delay(500);
  digitalWrite(PIN_FLASH_0, HIGH);
  delay(500);

  // clear buffer
  for (buffer_pnt = 0; buffer_pnt < BUFFER_SIZE; ++buffer_pnt) {
    flash_sum_buffer[buffer_pnt] = 0;
    flash_cnt_buffer[buffer_pnt] = 0;
  }; 
  buffer_pnt = 0;

  // start system
  set_sleep_mode(SLEEP_MODE_IDLE);
  sec = 0;
  sei();
}

// loop() ------------------------------------------------------------------------------------------------------------------------
void loop() {
  uint16_t max_flash_sum, avg_flash_sum;
  char str_buf[7] = " ";
    
  // check job flags
  if (job_flags & (1 << JF_SEC)) {                            // jobs "once per second"
    job_flags &= ~(1 << JF_SEC);

    // check whether the minute is completed
    if (sec == 0) {                                           // once per minute: update graph, calculate alarm level
      draw_graph(&max_flash_sum, &avg_flash_sum);
      // calculate and set alarm level
      alarm_level = calc_alarm_level(max_flash_sum, avg_flash_sum);
      if (alarm_level >= 1) display_cnt = DISPLAY_TIMEOUT;  
      if (alarm_level >= 2) digitalWrite(PIN_ALARM_2, LOW);
      else digitalWrite(PIN_ALARM_2, HIGH);
      if (alarm_level >= 3) digitalWrite(PIN_ALARM_3, LOW);
      else digitalWrite(PIN_ALARM_3, HIGH);
      if (alarm_level == 4) digitalWrite(PIN_ALARM_4, LOW);
      else digitalWrite(PIN_ALARM_4, HIGH);
    };

    // update current data (once per second) 
    gd.set_font(2);
    gd.set_cursor(DP_SEC_X, DP_SEC_Y);
    itoa(sec, str_buf + 1, 10);
    gd.draw_radj_str(str_buf);
    gd.set_cursor(DP_FLASH_CNT_X, DP_FLASH_CNT_Y);
    itoa(flash_cnt, str_buf + 1, 10);
    gd.draw_radj_str(str_buf);
    gd.set_cursor(DP_FLASH_SUM_X, DP_FLASH_SUM_Y);
    itoa(flash_sum, str_buf + 1, 10);
    gd.draw_radj_str(str_buf);
  
    // check and manage display backlight (once per second)
    if (display_cnt > 0) {
      if (display_status == 0) {
        gd.dim_off();
        display_status = 1;
      };
      if (--display_cnt == 0) {
        gd.dim_on();
        display_status = 0;
      };
    };
  }; // end of jobs "once per second"

  // run activities to be executed any time 
  // check keyboard
  if (keyboard_value > 0) {
    switch (keyboard_value) {
      case KEY_SELECT:             // select key
        break;
      case KEY_DOWN:             // down key
        if (graph_mode == 0) graph_mode = 3;
        else --graph_mode;
        draw_graph(&max_flash_sum, &avg_flash_sum);
        break;
      case KEY_UP:             // up key
        if (++graph_mode > 3) graph_mode = 0;
        draw_graph(&max_flash_sum, &avg_flash_sum);
        break;
      case KEY_CONFIG:            // configuration mode
        keyboard_value = 0;
        run_setup();
        draw_graph(&max_flash_sum, &avg_flash_sum);
        break;
    };
    keyboard_value = 0;
  }; // end keyboard check

  // sleep till next interrupt
  sleep_mode();
}

// ISR routine called by timer 1 interrupts ---------------------------------------------------------------------------------------
ISR(TIMER1_COMPA_vect){
  static uint16_t cnt = 0;
  static uint16_t old_adc_value_0 = 0;                        // previous value channel 0
  static uint16_t old_adc_value_1 = 0;                        // previous value channel 1
  static uint16_t adc_max_0 = 0, adc_max_1 = 0;               // maximum detection
  static uint16_t flash_led_cnt_0 = 0, flash_led_cnt_1 = 0;   // counter for shutoff of flash LEDs
  uint16_t adc_value;                                         // current adc value
  
  uint8_t key; 
  static uint8_t key_old = 0, key_cnt = 0;                    
  static uint8_t trigger_mode_0 = 0, trigger_mode_1 = 0;
  static uint8_t adc_channel = 0;                             // current adc channel (0 or 1)

  // read ADC and look for triggering edges
  adc_value = ADCW;                                     // capture current ADC value
  if (adc_channel == 0) {
    switch (trigger_mode_0) {
      case 0:                                             // wait for trigger condition
        if (adc_value > old_adc_value_0 + trigger_threshold) {
          trigger_mode_0 = 1;                            // set mode to "trigger detected"
          adc_max_0 = adc_value;                         // capture maximum
        }
        break;
      case 1:
        if (adc_value > adc_max_0) {                    // still raising edge?
          adc_max_0 = adc_value;                        // if yes, capture maximum
        } else {
          trigger_mode_0 = 0;                           // otherwise finish trigger mode
          if (flash_cnt < UPPER_LIMIT) ++flash_cnt;     // update flash counter
          if (flash_sum < UPPER_LIMIT)
            flash_sum += adc_max_0 / 10;                // update flash sum
          digitalWrite(PIN_FLASH_0, LOW);               // activate flash LED
          flash_led_cnt_0 = FLASH_LED_DUR;              // set flash LED duration
          adc_max_0 = 0;                                // clear maximum
        };
        break;
    };
    old_adc_value_0 = adc_value;                        // memorize current adc value for next round
    if (digitalRead(PIN_FLASH_0) == LOW) {              // check flash LED duration
      if (--flash_led_cnt_0 == 0)                       // shut off if needed
        digitalWrite(PIN_FLASH_0, HIGH);
    };
    adc_channel = 1;                                    // swap channel for the next round
  } else {
    switch (trigger_mode_1) {
      case 0:                                           // wait for trigger condition
        if (adc_value > old_adc_value_1 + trigger_threshold) {
          trigger_mode_1 = 1;                           // set mode to "trigger detected"
          adc_max_1 = adc_value;                        // capture maximum
        }
        break;
      case 1:
        if (adc_value > adc_max_1) {                    // still raising edge?
          adc_max_1 = adc_value;                        // if yes, capture maximum
        } else {
          trigger_mode_1 = 0;                           // otherwise finish trigger mode
          if (flash_cnt < UPPER_LIMIT) ++flash_cnt;     // update flash counter
          if (flash_sum < UPPER_LIMIT)
            flash_sum += adc_max_1 / 10;                // update flash sum
          digitalWrite(PIN_FLASH_1, LOW);               // activate flash LED
          flash_led_cnt_1 = FLASH_LED_DUR;              // set flash LED duration
          adc_max_1 = 0;                                // clear maximum
        };
        break;
    };
    old_adc_value_1 = adc_value;                        // memorize current adc value for next round
    if (digitalRead(PIN_FLASH_1) == LOW) {
      if (--flash_led_cnt_1 == 0)                       // check flash LED duration
        digitalWrite(PIN_FLASH_1, HIGH);                // shut off if needed
    };
    adc_channel = 0;                                    // swap channel for the next round
  };  
  ADMUX = (1 << REFS1) | adc_channel;                   // re-start ADC for the next round

  // check keyboard approximately 31 times per second
  if (cnt % 128 == 0) {
    key = 0;
    if (digitalRead(PIN_KEY_SELECT) == LOW) key = KEY_SELECT;
    if (digitalRead(PIN_KEY_DOWN) == LOW) key = KEY_DOWN;
    if (digitalRead(PIN_KEY_UP) == LOW) key = KEY_UP;
    if (key == 0) {                                     // no key pressed: nothing to do
        key_cnt = 0;
    } else {                                            // otherwise, analyze keyboard
      display_cnt = DISPLAY_TIMEOUT;                    // (re-) activate display backlight
      if (key != key_old) {                             // in case of any changes at the keyboard
        key_cnt = 0;                                    // reset key counter  
        key_old = key;                                  // remember new value
      } else {
        ++key_cnt;                                      // in case of stable keyboard data, count up
        switch (key_cnt) {                                
          case 3:                                       // after 3 times we assume stable keyboard data:
            keyboard_value = key;                       // capture pressed key
            break;
          case 15:                                      // after 15 times of stable keyboard data,
            if ((key == KEY_DOWN) || (key == KEY_UP)) { // for keys "up" and "down" 
              keyboard_value = key;                     // activate key repeat function
              key_cnt = 11;
            };
            break;
          case 93:                                      // for KEY_SELECT only: if the key is pressed ... 
            keyboard_value = KEY_CONFIG;                // ... for 3 seconds, activate configuration mode
            ++key_cnt;
            break;
          case 250:                                     // avoid overflow in case of continoulsy ...
            key_cnt = 249;                              // ... pressed key
            break;
        };
      };
    };
  }; 

  // manage seconds and minutes       
  if (++cnt == IRQ_FREQ) {                              // once per second
    job_flags |= (1 << JF_SEC);
    cnt = 0;
    if (++sec == 60) {                                  // every minute:
      if (++buffer_pnt == BUFFER_SIZE)                  // count up buffer pointer
        buffer_pnt = 0;
      flash_cnt_buffer[buffer_pnt] = flash_cnt;         // capture data to the buffer
      flash_sum_buffer[buffer_pnt] = flash_sum;
      flash_cnt = 0;                                    // reset all counters
      flash_sum = 0;
      sec = 0;
    };
  };
}

